////////////////////////////////////////////////////////////
// WinSock debug utility								  //
// (C) 2002 by Thomas Bleeker							  //
// www.MadWizard.org									  //
////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "psapi.h"
#include "common.h"

typedef unsigned long u32;
typedef unsigned short u16;
typedef unsigned char u8;



typedef struct
{
	char*	lpLibrary;		// DLL name
	char*	lpFunction;		// function name
	u32		ordinal;		// Ordinal of function
	u32		RVA;			// RVA of entry in FirstThunk table
} API_HOOK_INFO;


API_HOOK_INFO hookTable[] = 
{
	{"ws2_32.dll","socket",0x17,0},
	{"ws2_32.dll","recv",0x10,0},
	{"ws2_32.dll","send",0x13,0},
	{"ws2_32.dll","closesocket",0x03,0},
	{"ws2_32.dll","accept",0x01,0},
	{"ws2_32.dll","WSASocketA",0x4E,0},
	{"ws2_32.dll","WSASocketW",0x4F,0},
	{"ws2_32.dll","WSASend",0x48,0},
	{"ws2_32.dll","WSAAccept",0x1A,0},
	{"ws2_32.dll","WSARecv",0x43,0},
	
	{"wsock32.dll","accept",0x01,0},
	{"wsock32.dll","closesocket",0x03,0},
	{"wsock32.dll","recv",0x10,0},
	{"wsock32.dll","send",0x13,0},
	{"wsock32.dll","socket",0x17,0}

};


/*
	Hook procedure:

	58      pop       eax		// pop return address off stack
	CC      int       3			// first break, before call
	FFD0    call      eax		// call API
	CC		int		  3			// second break, after call
*/

#pragma pack(1)
typedef struct
{
	// Actual code:
	u8	  popEAX;		// 0x58 ; pop eax ; pop return address
	u8	  break1;		// 0xCC ; int 3 ; breakpoint 1
	u16	  callEAX;		// 0xD0FF ; call eax ; call old function
	u8	  break2;		// 0xCC ; int 3 ; breakpoint 2.
	
	// Stored info (but copied to process mem as well):
	u32	 pOldFunc;		// Pointer to old function
	u8   padding[5];	// Padding to 16 bytes alignment
} HOOKMEMITEM;
#pragma pack()



extern void HandleHook(HANDLE hProcess, ACTIVE_HOOK *activeHook, CONTEXT *context);


const int MAX_SIMULTANIOUS_HOOKS = 50;

HOOKMEMITEM localHookMem[arrcount(hookTable)];

ACTIVE_HOOK activeHooks[MAX_SIMULTANIOUS_HOOKS];


typedef void (__stdcall * IMPORTCALLBACK)(LPSTR lpLibrary, 
							  LPSTR lpFunction,
							  u32	ordinal,
							  u32   dwRVA);
// lpLibrary: DLL name string
// lpFunction: function name string, if NULL, ordinal is valid
// ordinal: ordinal of function, only valid if lpFunction is NULL.


bool EnumImports(LPSTR lpImageFile,
				 IMPORTCALLBACK lpCallBackFunc)
{
	HANDLE	hFile, hMap;
	void	*pMap, *curPtr;

	// Load DLL
	if (!(hFile = CreateFile(lpImageFile,GENERIC_READ,
				FILE_SHARE_READ|FILE_SHARE_DELETE,NULL,OPEN_EXISTING,NULL,NULL)))
		return false;

	// Create file mapping
	if (!(hMap = CreateFileMapping(hFile,NULL,PAGE_READONLY|SEC_COMMIT,
				0,0,NULL)))
	{
		CloseHandle(hFile);
		return false;
	}

	if(!(pMap = MapViewOfFile(hMap,FILE_MAP_READ,0,0,0)))
	{
		CloseHandle(hFile);
		CloseHandle(hMap);
		return false;
	}

	curPtr = pMap;

	if (((IMAGE_DOS_HEADER*)pMap)->e_magic!=IMAGE_DOS_SIGNATURE)
	{
		MessageBeep(0);
	}

	curPtr = (void*)((long)curPtr + ((IMAGE_DOS_HEADER*)pMap)->e_lfanew);

	if (((IMAGE_NT_HEADERS*)curPtr)->Signature!=IMAGE_NT_SIGNATURE)
	{
		MessageBeep(0);
	}

	//	curPtr points to NT header
	
	IMAGE_IMPORT_DESCRIPTOR *pImportDescriptor;

	pImportDescriptor = (IMAGE_IMPORT_DESCRIPTOR*)
						RVAToOffset(pMap,
						((IMAGE_NT_HEADERS*)curPtr)->OptionalHeader.DataDirectory[1].VirtualAddress);
	if (pImportDescriptor==0)
	{
		UnmapViewOfFile(pMap);
		CloseHandle(hMap);
		CloseHandle(hFile);
		return true;  // no import table, no imports 
	}

	pImportDescriptor = (IMAGE_IMPORT_DESCRIPTOR*)((long)pImportDescriptor + (long)pMap);


	while(!(pImportDescriptor->OriginalFirstThunk==0 &&
			pImportDescriptor->TimeDateStamp==0 &&
			pImportDescriptor->ForwarderChain==0 &&
			pImportDescriptor->Name==0 &&
			pImportDescriptor->FirstThunk==0))
	{
		//DLL name:
		char *curDLLName = (char*)((long)pMap + (long)RVAToOffset(pMap, pImportDescriptor->Name));
		PIMAGE_THUNK_DATA pThunkData = (pImportDescriptor->OriginalFirstThunk==0) ?
						(PIMAGE_THUNK_DATA)pImportDescriptor->FirstThunk :
						(PIMAGE_THUNK_DATA)pImportDescriptor->OriginalFirstThunk;
		
		pThunkData = (PIMAGE_THUNK_DATA)
					 (RVAToOffset(pMap,(long)pThunkData) + (long)pMap);

		
		// curImportRVA is the RVA of the DWORD that contains the
		// address of the imported function when the image is loaded.
		
		u32 curImportRVA = pImportDescriptor->FirstThunk;


		while(pThunkData->u1.Ordinal!=0)
		{
			if (pThunkData->u1.Ordinal&IMAGE_ORDINAL_FLAG32)
			{
				//import by ordinal
				//ordinal is 
				lpCallBackFunc(curDLLName, NULL, pThunkData->u1.Ordinal&(0xFFFF),curImportRVA);
			}
			else
			{
				char *funcName = (char*)((long)pMap + RVAToOffset(pMap,(long)((PIMAGE_IMPORT_BY_NAME)(pThunkData->u1.AddressOfData))->Name));
				lpCallBackFunc(curDLLName, funcName, 0,curImportRVA);
				
				//MessageBox(0,(char*)
				//	((long)pMap + RVAToOffset(pMap,(long)pThunkData->u1.AddressOfData->Name)),0,0);
			}
			pThunkData++;
			
			curImportRVA += 4;
		}

		pImportDescriptor++;
	}



	UnmapViewOfFile(pMap);
	CloseHandle(hMap);

	CloseHandle(hFile);
	

	return true;
}


void CALLBACK ImportCallBack(LPSTR lpLibrary, LPSTR lpFunction, u32 ordinal, u32 dwRVA)
{
	int i;

	
	for (i=0;i<arrcount(hookTable);i++)
	{
		if (lpFunction!=0) // import by name
		{
			// See if name & DLL name matches
			if (lstrcmp(hookTable[i].lpFunction, lpFunction)==0 &&
				lstrcmpi(hookTable[i].lpLibrary, lpLibrary)==0)
			{
				// We got a match
		//		printf("match: %s\n", lpFunction);
				hookTable[i].RVA = dwRVA;
			}
		}
		else	// import by ordinal
		{
			if (hookTable[i].ordinal==ordinal &&
				lstrcmpi(hookTable[i].lpLibrary, lpLibrary)==0)
			{
				// We got a match
			//	printf("match: 0x%X\n", ordinal);
				hookTable[i].RVA = dwRVA;

			}
		}
	}
}

STARTUPINFO siInfo;
PROCESS_INFORMATION piInfo;


void *pRemoteHookMem;


void OnProgramStartup()
{
	
    HMODULE hMods[1024];
    u32		cbNeeded;
    unsigned int i;
	u32		bytesRead;

	HANDLE hMainProcess;

	// Open process with read & write access
	hMainProcess = OpenProcess(PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION,
								false, piInfo.dwProcessId);

	// Allocate memory in debuggee process to write
	// hook code to.
	pRemoteHookMem = VirtualAllocEx(piInfo.hProcess, NULL, sizeof(localHookMem),
						MEM_COMMIT, PAGE_EXECUTE_READWRITE);

	// Enumerate all modules in debuggee
    if(EnumProcessModules(piInfo.hProcess, hMods, sizeof(hMods), &cbNeeded))
    {
		// Loop through all modules
        for ( i = 0; i < (cbNeeded / sizeof(HMODULE)); i++ )
        {
			// buffer to store module filename
            char szModName[MAX_PATH];

            // Get the full path to the module's file.
			if ( GetModuleFileNameEx( piInfo.hProcess, hMods[i], szModName,
                                      sizeof(szModName)))
            {
                MODULEINFO modInfo;
			
				// Get more info (baseaddress etc):
				GetModuleInformation(piInfo.hProcess, hMods[i], &modInfo, sizeof(modInfo));

				// Clear RVAs in table. Some RVAs will stay NULL on return
				// of EnumImports indicating that function is not in the
				// import table. The RVAs have to be initialized to zero 
				// before calling EnumImports
				for (int k=0;k<arrcount(hookTable);k++)
					hookTable[k].RVA = 0;
				
				// Enumerate imports for this module:
				EnumImports(szModName, &ImportCallBack);
				
				// Loop through modified hookTable:
				for (k=0;k<arrcount(hookTable);k++)
				{
					if (hookTable[k].RVA!=0)
					{
				
						void	*temp;
						u32		oldProtection1, oldProtection2;

						// adresss of dword containing the RVA of the function
						void	*addressRVA = (void*) 
								((long)modInfo.lpBaseOfDll + (long)hookTable[k].RVA);

						// Use temp as 4-byte buffer:
						temp = (void*)((long)pRemoteHookMem + (k * sizeof(localHookMem[0])));
						
						// Read memory using the pOldFunc member of the items
						// in localHookMem as buffer. This member will contain
						// the old address of the function:
						ReadProcessMemory(hMainProcess,addressRVA,
								&localHookMem[k].pOldFunc, 4, &bytesRead);
						
						// Unprotect pages to be able to write to it:
						VirtualProtectEx(hMainProcess, (void*)((long)addressRVA & 0xFFFFF000),
										4096,PAGE_EXECUTE_READWRITE, &oldProtection1);

						// Overwrite function address with hook address:
						WriteProcessMemory(hMainProcess,addressRVA,&temp,4,
										&bytesRead);

						// Restore protection:
						VirtualProtectEx(hMainProcess, (void*)((long)addressRVA & 0xFFFFF000),
										4096,oldProtection1, &oldProtection2);

						// Set members of localHookMem item to form a 
						// piece of code:
						localHookMem[k].break1 = 0xCC;
						localHookMem[k].break2 = 0xCC;
						localHookMem[k].callEAX = 0xD0FF;
						localHookMem[k].popEAX = 0x58;
					}
				}

				// Write modified localHookMem table to remoteHookMem table
				WriteProcessMemory(hMainProcess,pRemoteHookMem,&localHookMem,sizeof(localHookMem),
							&bytesRead);
            }
        }
    }
	// Close opened handle to process
	CloseHandle(hMainProcess);
}


ACTIVE_HOOK* activateHook()
{
	for (int i=0;i<arrcount(activeHooks);i++)
	{
		if (!activeHooks[i].bActive)
		{
			activeHooks[i].bActive = true;
			return &activeHooks[i];
		}
	}
	return 0;
}



bool hookBreak(u32 dwThreadID)
{
	ACTIVE_HOOK* activeHook;
		
	CONTEXT		context;
	u32			bytesRead;
	
	HANDLE hThread = OpenThread(THREAD_GET_CONTEXT|THREAD_SET_CONTEXT,
			   false,
			   dwThreadID);

	context.ContextFlags = CONTEXT_INTEGER|CONTEXT_CONTROL;
	GetThreadContext(hThread, &context);	
	
	// EAX is return address
	/*
	58      pop       eax
	CC      int       3  <---------- here now
	FFD0    call      eax
	CC      int       3
	*/

	if (((long)context.Eip - (long)pRemoteHookMem) >= sizeof(localHookMem))
	{
		//real breakpoint, not in hook mem
		return false;
	}

	int hookMemItemOffset =  ((long)context.Eip - (long)pRemoteHookMem) %
							sizeof(HOOKMEMITEM);
	if (hookMemItemOffset==2) // first int 3 at relative offset 1 (2 because it has executed the instruction)
	{
		activeHook = activateHook();
		if (activeHook==0)
		{
			MessageBox(NULL, 
				"Error: maximum number of simultaniously "
				"active hooks reached. ", "wshook", MB_ICONERROR);
			return false;
		}
		activeHook->pReturnAddress = context.Eax;
		activeHook->hookIndex = ((long)context.Eip - (long)pRemoteHookMem) /
								sizeof(HOOKMEMITEM);
		activeHook->oldEDI = context.Edi; // save EDI because it's overwritten
		context.Edi = (unsigned long)activeHook;
		context.Eax = localHookMem[activeHook->hookIndex].pOldFunc; 
		SetThreadContext(hThread, &context); // set new EDI & EAX
		ReadProcessMemory(piInfo.hProcess, (void*)context.Esp, &activeHook->params[0], 12*4,
						&bytesRead);
	//	printf("hook: %08X [%d]\n",context.Eip, activeHook->hookIndex);
	}
	else	// second int 3 (after function call)
	{
		activeHook = (ACTIVE_HOOK*)context.Edi;
		context.Edi = activeHook->oldEDI;
		HandleHook(piInfo.hProcess,activeHook, &context);
		context.Eip = activeHook->pReturnAddress;
		SetThreadContext(hThread,&context);
		activeHook->bActive = false;
	}
	CloseHandle(hThread);
	return true;
}


u32 WINAPI DebugThread(LPVOID lpParameter)
{
	
	
	bool	bStartupBreak = true;	// True if int3 breakpoint
									// is first break on program startup
	

	// Initialize startup info
	siInfo.cb = sizeof(siInfo);
	GetStartupInfo(&siInfo);

	// Create process as debuggee:
	if (!CreateProcess((char*)lpParameter,0,
			NULL, NULL,TRUE,DEBUG_PROCESS|DEBUG_ONLY_THIS_PROCESS,
			NULL, NULL,
			&siInfo, &piInfo))
	{
		MessageBox(0,"Couldn't create process","debug",0);
		return false;
	}
	

	DEBUG_EVENT	dbEvent;

	// Inifinite loop while debuggee runs
	while(true)
	{
		// Wait for debug event to happen
		WaitForDebugEvent(&dbEvent, INFINITE);

		// Exit? break out of infinite loop
		if(dbEvent.dwDebugEventCode==EXIT_PROCESS_DEBUG_EVENT)
		{
			if (dbEvent.dwProcessId==piInfo.dwProcessId)
				break;
		}
		// Loading DLL?
		else if(dbEvent.dwDebugEventCode==LOAD_DLL_DEBUG_EVENT)
		{
			// not used (yet)
		}
		// Exception?
		else if(dbEvent.dwDebugEventCode==EXCEPTION_DEBUG_EVENT)
		{
			// Int 3 exception?
			if (dbEvent.u.Exception.ExceptionRecord.ExceptionCode==EXCEPTION_BREAKPOINT)
			{
				// If first break, it's the default breakpoint that
				// occurs on program startup
				if (bStartupBreak)
				{
					OnProgramStartup();					
					bStartupBreak = false;
					ContinueDebugEvent(dbEvent.dwProcessId, dbEvent.dwThreadId, DBG_CONTINUE);
				}
				// Otherwise, it's a breakpoint in the hook routine:
				else
				{
					// Handle break
					hookBreak(dbEvent.dwThreadId);
					ContinueDebugEvent(dbEvent.dwProcessId, dbEvent.dwThreadId, DBG_CONTINUE);
					continue;
				}
			}
		}
		ContinueDebugEvent(dbEvent.dwProcessId, dbEvent.dwThreadId, DBG_EXCEPTION_NOT_HANDLED);
	}
	
	// Free memory allocated for hook routines
	VirtualFreeEx(piInfo.hProcess, pRemoteHookMem, 0, MEM_RELEASE);

	// Close handles
	CloseHandle(piInfo.hProcess);
	CloseHandle(piInfo.hThread);
	
	return true;
}




